var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _possibleConstructorReturn(self, call) { if (!self) { throw new ReferenceError("this hasn't been initialised - super() hasn't been called"); } return call && (typeof call === "object" || typeof call === "function") ? call : self; }

function _inherits(subClass, superClass) { if (typeof superClass !== "function" && superClass !== null) { throw new TypeError("Super expression must either be null or a function, not " + typeof superClass); } subClass.prototype = Object.create(superClass && superClass.prototype, { constructor: { value: subClass, enumerable: false, writable: true, configurable: true } }); if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass; }

import React from 'react';
import PropTypes from 'prop-types';
import { getRequestKey, fetchDedupe, isRequestInFlight } from 'fetch-dedupe';

// This object is our cache
// The keys of the object are requestKeys
// The value of each key is a Response instance
var responseCache = {};

export function clearResponseCache() {
  responseCache = {};
}

export var Fetch = function (_React$Component) {
  _inherits(Fetch, _React$Component);

  _createClass(Fetch, [{
    key: 'render',
    value: function render() {
      var _props = this.props,
          children = _props.children,
          requestName = _props.requestName;
      var _state = this.state,
          fetching = _state.fetching,
          response = _state.response,
          data = _state.data,
          error = _state.error,
          requestKey = _state.requestKey,
          url = _state.url;


      if (!children) {
        return null;
      } else {
        return children({
          requestName: requestName,
          url: url,
          fetching: fetching,
          failed: Boolean(error || response && !response.ok),
          response: response,
          data: data,
          requestKey: requestKey,
          error: error,
          doFetch: this.fetchRenderProp
        }) || null;
      }
    }
  }]);

  function Fetch(props, context) {
    _classCallCheck(this, Fetch);

    var _this = _possibleConstructorReturn(this, (Fetch.__proto__ || Object.getPrototypeOf(Fetch)).call(this, props, context));

    _initialiseProps.call(_this);

    _this.state = {
      requestKey: props.requestKey || getRequestKey(_extends({}, props, {
        method: props.method.toUpperCase()
      })),
      requestName: props.requestName,
      fetching: false,
      response: null,
      data: null,
      error: null,
      url: props.url || ''
    };
    return _this;
  }

  // We default to being lazy for "write" requests,
  // such as POST, PATCH, DELETE, and so on.


  _createClass(Fetch, [{
    key: 'componentDidMount',
    value: function componentDidMount() {
      if (!this.isLazy()) {
        this.fetchData();
      }
    }
  }, {
    key: 'componentDidUpdate',
    value: function componentDidUpdate(prevProps) {
      var currentRequestKey = this.props.requestKey || getRequestKey(_extends({}, this.props, {
        method: this.props.method.toUpperCase()
      }));
      //
      var prevRequestKey = prevProps.requestKey || getRequestKey(_extends({}, prevProps, {
        method: prevProps.method.toUpperCase()
      }));

      if (currentRequestKey !== prevRequestKey && !this.isLazy(prevProps)) {
        this.fetchData({
          requestKey: currentRequestKey
        });
      }
    }
  }, {
    key: 'componentWillUnmount',
    value: function componentWillUnmount() {
      this.willUnmount = true;
      this.cancelExistingRequest('Component unmounted');
    }

    // When a request is already in flight, and a new one is
    // configured, then we need to "cancel" the previous one.


    // When a subsequent request is made, it is important that the correct
    // request key is used. This method computes the right key based on the
    // options and props.


    // Heads up: accessing `this.props` within `fetchData`
    // is unsafe! This is because we call `fetchData` from
    // within `componentWillReceiveProps`!

  }]);

  return Fetch;
}(React.Component);

var _initialiseProps = function _initialiseProps() {
  var _this2 = this;

  this.isReadRequest = function (method) {
    var uppercaseMethod = method.toUpperCase();

    return uppercaseMethod === 'GET' || uppercaseMethod === 'HEAD' || uppercaseMethod === 'OPTIONS';
  };

  this.isLazy = function (props) {
    var _ref = props || _this2.props,
        lazy = _ref.lazy,
        method = _ref.method;

    return typeof lazy === 'undefined' ? !_this2.isReadRequest(method) : lazy;
  };

  this.shouldCacheResponse = function () {
    var _props2 = _this2.props,
        cacheResponse = _props2.cacheResponse,
        method = _props2.method;


    return typeof cacheResponse === 'undefined' ? _this2.isReadRequest(method) : cacheResponse;
  };

  this.getFetchPolicy = function () {
    var _props3 = _this2.props,
        fetchPolicy = _props3.fetchPolicy,
        method = _props3.method;


    if (typeof fetchPolicy === 'undefined') {
      return _this2.isReadRequest(method) ? 'cache-first' : 'network-only';
    } else {
      return fetchPolicy;
    }
  };

  this.cancelExistingRequest = function (reason) {
    if (_this2.state.fetching && !_this2.hasHandledNetworkResponse) {
      var abortError = new Error(reason);
      abortError.name = 'AbortError';
      _this2.onResponseReceived(_extends({}, _this2.responseReceivedInfo, {
        error: abortError,
        hittingNetwork: true
      }));
    }
  };

  this.fetchRenderProp = function (options) {
    // We wrap this in a setTimeout so as to avoid calls to `setState`
    // in render, which React does not allow.
    //
    // tl;dr, the following code should never cause a problem:
    //
    // `<Fetch children={({ doFetch }) => doFetch()} />
    setTimeout(function () {
      _this2.fetchData(options, true);
    });
  };

  this.getRequestKey = function (options) {
    // A request key in the options gets top priority
    if (options && options.requestKey) {
      return options.requestKey;
    }

    // Otherwise, if we have no request key, but we do have options, then we
    // recompute the request key based on these options.
    // Note that if the URL, body, or method have not changed, then the request
    // key should match the previous request key if it was computed.
    // If you passed in a custom request key as a prop, then you will also
    // need to pass in a custom key when you call `doFetch()`!
    else if (options) {
        var _Object$assign = Object.assign({}, _this2.props, options),
            url = _Object$assign.url,
            method = _Object$assign.method,
            body = _Object$assign.body;

        return getRequestKey({
          url: url,
          body: body,
          method: method.toUpperCase()
        });
      }

      // Next in line is the the request key from props.
      else if (_this2.props.requestKey) {
          return _this2.props.requestKey;
        }

        // Lastly, we compute the request key from the props.
        else {
            var _props4 = _this2.props,
                _url = _props4.url,
                _method = _props4.method,
                _body = _props4.body;


            return getRequestKey({
              url: _url,
              body: _body,
              method: _method.toUpperCase()
            });
          }
  };

  this.fetchData = function (options, ignoreCache) {
    // These are the things that we do not allow a user to configure in
    // `options` when calling `doFetch()`. Perhaps we should, however.
    var _props5 = _this2.props,
        requestName = _props5.requestName,
        dedupe = _props5.dedupe,
        beforeFetch = _props5.beforeFetch;


    _this2.cancelExistingRequest('New fetch initiated');

    var requestKey = _this2.getRequestKey(options);
    var requestOptions = Object.assign({}, _this2.props, options);

    var url = requestOptions.url,
        body = requestOptions.body,
        credentials = requestOptions.credentials,
        headers = requestOptions.headers,
        method = requestOptions.method,
        responseType = requestOptions.responseType,
        mode = requestOptions.mode,
        cache = requestOptions.cache,
        redirect = requestOptions.redirect,
        referrer = requestOptions.referrer,
        referrerPolicy = requestOptions.referrerPolicy,
        integrity = requestOptions.integrity,
        keepalive = requestOptions.keepalive,
        signal = requestOptions.signal;


    var uppercaseMethod = method.toUpperCase();
    var shouldCacheResponse = _this2.shouldCacheResponse();

    var init = {
      body: body,
      credentials: credentials,
      headers: headers,
      method: uppercaseMethod,
      mode: mode,
      cache: cache,
      redirect: redirect,
      referrer: referrer,
      referrerPolicy: referrerPolicy,
      integrity: integrity,
      keepalive: keepalive,
      signal: signal
    };

    var responseReceivedInfo = {
      url: url,
      init: init,
      requestKey: requestKey,
      responseType: responseType
    };

    // This is necessary because `options` may have overridden the props.
    // If the request config changes, we need to be able to accurately
    // cancel the in-flight request.
    _this2.responseReceivedInfo = responseReceivedInfo;

    _this2.hasHandledNetworkResponse = false;

    var fetchPolicy = _this2.getFetchPolicy();

    var cachedResponse = void 0;
    if (fetchPolicy !== 'network-only' && !ignoreCache) {
      cachedResponse = responseCache[requestKey];

      if (cachedResponse) {
        _this2.onResponseReceived(_extends({}, responseReceivedInfo, {
          response: cachedResponse,
          hittingNetwork: false,
          stillFetching: fetchPolicy === 'cache-and-network'
        }));

        if (fetchPolicy === 'cache-first' || fetchPolicy === 'cache-only') {
          return Promise.resolve(cachedResponse);
        }
      } else if (fetchPolicy === 'cache-only') {
        var cacheError = new Error('Response for "' + requestName + '" not found in cache.');
        _this2.onResponseReceived(_extends({}, responseReceivedInfo, {
          error: cacheError,
          hittingNetwork: false
        }));
        return Promise.resolve(cacheError);
      }
    }

    _this2.setState({
      // Note: when data is returned from the cache, the calls to `onResponseReceived`
      // will call `setState` to update the `requestKey`. This call to update `requestKey`
      // may be duplicated in those situations, but that should be OK. It is necessary
      // to include this here to account for "network-only" fetch policies.
      requestKey: requestKey,
      url: url,
      error: null,
      failed: false,
      fetching: true
    });
    var hittingNetwork = !isRequestInFlight(requestKey) || !dedupe;

    if (hittingNetwork) {
      beforeFetch({
        url: url,
        init: init,
        requestKey: requestKey
      });
    }
    return fetchDedupe(url, init, { requestKey: requestKey, responseType: responseType, dedupe: dedupe }).then(function (res) {
      if (shouldCacheResponse) {
        responseCache[requestKey] = res;
      }

      if (!_this2.hasHandledNetworkResponse) {
        _this2.onResponseReceived(_extends({}, responseReceivedInfo, {
          response: res,
          hittingNetwork: hittingNetwork
        }));
      }

      return res;
    }, function (error) {
      if (!_this2.hasHandledNetworkResponse) {
        _this2.onResponseReceived(_extends({}, responseReceivedInfo, {
          error: error,
          cachedResponse: cachedResponse,
          hittingNetwork: hittingNetwork
        }));
      }

      return error;
    });
  };

  this.onResponseReceived = function (info) {
    var _info$error = info.error,
        error = _info$error === undefined ? null : _info$error,
        _info$response = info.response,
        response = _info$response === undefined ? null : _info$response,
        hittingNetwork = info.hittingNetwork,
        url = info.url,
        init = info.init,
        requestKey = info.requestKey,
        cachedResponse = info.cachedResponse,
        _info$stillFetching = info.stillFetching,
        stillFetching = _info$stillFetching === undefined ? false : _info$stillFetching;


    _this2.responseReceivedInfo = null;

    if (!stillFetching) {
      _this2.hasHandledNetworkResponse = true;
    }

    var data = void 0;
    // If our response succeeded, then we use that data.
    if (response && response.data) {
      data = response.data;
    } else if (cachedResponse && cachedResponse.data) {
      // This happens when the request failed, but we have cache-and-network
      // specified. Although we pass along the failed response, we continue to
      // pass in the cached data.
      data = cachedResponse.data;
    }

    data = data ? _this2.props.transformData(data) : null;

    if (hittingNetwork) {
      _this2.props.afterFetch({
        url: url,
        init: init,
        requestKey: requestKey,
        error: error,
        failed: Boolean(error || response && !response.ok),
        response: response,
        data: data,
        didUnmount: Boolean(_this2.willUnmount)
      });
    }

    if (_this2.willUnmount) {
      return;
    }

    _this2.setState({
      url: url,
      data: data,
      error: error,
      response: response,
      fetching: stillFetching,
      requestKey: requestKey
    }, function () {
      return _this2.props.onResponse(error, response);
    });
  };
};

var globalObj = typeof self !== 'undefined' ? self : this;
var AbortSignalCtr = globalObj.AbortSignal || function () {};

Fetch.propTypes = {
  children: PropTypes.func,
  requestName: PropTypes.string,
  fetchPolicy: PropTypes.oneOf(['cache-first', 'cache-and-network', 'network-only', 'cache-only']),
  onResponse: PropTypes.func,
  beforeFetch: PropTypes.func,
  afterFetch: PropTypes.func,
  responseType: PropTypes.oneOfType([PropTypes.func, PropTypes.oneOf(['json', 'text', 'blob', 'arrayBuffer', 'formData'])]),
  transformData: PropTypes.func,
  lazy: PropTypes.bool,
  dedupe: PropTypes.bool,
  requestKey: PropTypes.string,

  url: PropTypes.string.isRequired,
  body: PropTypes.any,
  credentials: PropTypes.oneOf(['omit', 'same-origin', 'include']),
  headers: PropTypes.object,
  method: PropTypes.oneOf(['get', 'post', 'put', 'patch', 'delete', 'options', 'head', 'GET', 'POST', 'PUT', 'PATCH', 'DELETE', 'OPTIONS', 'HEAD']),
  mode: PropTypes.oneOf(['same-origin', 'cors', 'no-cors', 'navigate', 'websocket']),
  cache: PropTypes.oneOf(['default', 'no-store', 'reload', 'no-cache', 'force-cache', 'only-if-cached']),
  redirect: PropTypes.oneOf(['manual', 'follow', 'error']),
  referrer: PropTypes.string,
  referrerPolicy: PropTypes.oneOf(['no-referrer', 'no-referrer-when-downgrade', 'origin', 'origin-when-cross-origin', 'unsafe-url', '']),
  integrity: PropTypes.string,
  keepalive: PropTypes.bool,
  signal: PropTypes.instanceOf(AbortSignalCtr)
};

Fetch.defaultProps = {
  requestName: 'anonymousRequest',
  onResponse: function onResponse() {},
  beforeFetch: function beforeFetch() {},
  afterFetch: function afterFetch() {},
  transformData: function transformData(data) {
    return data;
  },
  dedupe: true,

  method: 'get',
  referrerPolicy: '',
  integrity: '',
  referrer: 'about:client'
};